/* Shared library add-on to iptables to add sprio matching support. 
 * 
 * Copyright (C) 2000 Emmanuel Roger  <winfield@freegates.be>
 *
 * ChangeLog
 *     27.01.2001: Gianni Tedesco <gianni@ecsc.co.uk>
 *             Changed --tos to --sprio in save(). Also
 *             updated to work with slightly modified
 *             ipt_sprio_info.
 */

 /*
 SERCOMM   Supirior Prio  2007.05.21
-m  lanport --port [1~n]

 */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <ctype.h>

#include <iptables.h>
#include "../../../apps/fw_modules_priv/ipt_ifmap.h"

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"Interface Port match v%s options:\n"
"-m ifmap --type [1:lan 2:wifi] --port [1~n]       Match skb->source_ifport in a packet\n",
IPTABLES_VERSION);

	fputc('\n', stdout);
}

/*
{ "port", 1, 0, '1' }
1st param:
  flag name
  
2nd param:
  0: no value follow the --param
  1: must have a value
  2: optional

4th param:
  it's more like an ID for --param;   this value will be used in  switch(c)   and   case'1'
  we use a char '1', which is in fact stored in corresponding int value.  
  this int will be returned as in  parse(int c,...)
*/
static struct option opts[] = { 
	{ "type", 1, 0, '1' },
	{ "port", 1, 0, '2' },
	{0}
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	*nfcache |= NFC_UNKNOWN;
}

static void
parse_iftype(const unsigned char *p, struct ipt_ifmap_info *info)
{	
	u_int32_t lan =0x00010000;
	u_int32_t wifi=0x00020000;

	if (strlen(p) == 1){
		switch (*p){
		case '1'://lan
			info->source_iftype = lan;
			break;
		case '2'://wifi
			info->source_iftype = wifi;
			break;
		}
	}
	else{
		exit_error(PARAMETER_PROBLEM, "invalid interface port setting `%s' specified", p);
	}
	
}


static void
parse_port(const unsigned char *p, struct ipt_ifmap_info *info)
{	
	u_int32_t port_id;
	if (string_to_number(p, 0, 64, &port_id) != -1){ //it's very unlikely to have more than 64 ports!
		info->source_ifport = port_id;
	}
	else{
		exit_error(PARAMETER_PROBLEM, "invalid interface port setting `%s' specified", p);
	}
	
}


/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_ifmap_info *ifmapinfo = (struct ipt_ifmap_info *)(*match)->data;

	switch (c) {
	case '1':
		check_inverse(optarg, &invert, &optind, 0);
		parse_iftype(argv[optind-1], ifmapinfo);
		if (invert)
			ifmapinfo->invert = 1;
		*flags |= 0x0001;
		break;

	case '2':
		check_inverse(optarg, &invert, &optind, 0);
		parse_port(argv[optind-1], ifmapinfo);
		if (invert)
			ifmapinfo->invert = 1;
		*flags |= 0x0010;
		break;


	default:
		return 0;
	}
	return 1;
}

static void
print_port(u_int32_t ifport, u_int32_t iftype, int invert)
{
	if (invert)
		fputc('!', stdout);
	printf("port:%u    type:%u", ifport, iftype>>16);
}

/* Final check; must have specified --port. */
static void
final_check(unsigned int flags)
{
	if (!(flags & 0x0001))
		exit_error(PARAMETER_PROBLEM,
			   "Interface Port match: You must specify `--type");
	if (!(flags & 0x0010))
		exit_error(PARAMETER_PROBLEM,
			   "Interface Port match: You must specify `--port");

}

/* Prints out the matchinfo. */
static void print(const struct ipt_ip *ip,      const struct ipt_entry_match *match,      int numeric){
	print_port( ((struct ipt_ifmap_info *)match->data)->source_ifport, ((struct ipt_ifmap_info *)match->data)->source_iftype,  ((struct ipt_ifmap_info *)match->data)->invert  );
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	printf("--port ");
	print_port( ((struct ipt_ifmap_info *)match->data)->source_ifport, ((struct ipt_ifmap_info *)match->data)->source_iftype,  ((struct ipt_ifmap_info *)match->data)->invert  );}

static
struct iptables_match ifmap_match
= { NULL,
    "ifmap",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_ifmap_info)),
    IPT_ALIGN(sizeof(struct ipt_ifmap_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&ifmap_match);
}

