/* ==========================================================================
 * $File: //dwh/usb_iip/dev/software/otg_ipmate/linux/drivers/dwc_otg_driver.c $
 * $Revision: #6 $
 * $Date: 2006/04/21 $
 * $Change: 631780 $
 *
 * Synopsys HS OTG Linux Software Driver and documentation (hereinafter,
 * "Software") is an Unsupported proprietary work of Synopsys, Inc. unless
 * otherwise expressly agreed to in writing between Synopsys and you.
 *
 * The Software IS NOT an item of Licensed Software or Licensed Product under
 * any End User Software License Agreement or Agreement for Licensed Product
 * with Synopsys or any supplement thereto. You are permitted to use and
 * redistribute this Software in source and binary forms, with or without
 * modification, provided that redistributions of source code must retain this
 * notice. You may not view, use, disclose, copy or distribute this file or
 * any information contained herein except pursuant to this license grant from
 * Synopsys. If you do not agree with this notice, including the disclaimer
 * below, then you are not authorized to use the Software.
 *
 * THIS SOFTWARE IS BEING DISTRIBUTED BY SYNOPSYS SOLELY ON AN "AS IS" BASIS
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE HEREBY DISCLAIMED. IN NO EVENT SHALL SYNOPSYS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * ========================================================================== */

/** @file
 * The dwc_otg_driver module provides the initialization and cleanup entry
 * points for the DWC_otg driver. This module will be dynamically installed
 * after Linux is booted using the insmod command. When the module is
 * installed, the dwc_otg_driver_init function is called. When the module is
 * removed (using rmmod), the dwc_otg_driver_cleanup function is called.
 *
 * This module also defines a data structure for the dwc_otg_driver, which is
 * used in conjunction with the standard ARM lm_device structure. These
 * structures allow the OTG driver to comply with the standard Linux driver
 * model in which devices and drivers are registered with a bus driver. This
 * has the benefit that Linux can expose attributes of the driver and device
 * in its special sysfs file system. Users can then read or write files in
 * this file system to perform diagnostics on the driver components or the
 * device.
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#ifndef CONFIG_IFX // winder
#include <linux/device.h>
#endif

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/stat.h>  /* permission constants */

#include <asm/io.h>

// 20070316, winder.
#ifndef CONFIG_IFX
#include <asm/arch/lm.h>
#include <asm/sizes.h>
#endif

#ifndef CONFIG_IFX // winder
#include "linux/dwc_otg_plat.h"
#else
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include "dwc_otg_plat.h"
#endif

#ifndef CONFIG_IFX
#include "dwc_otg_attr.h"
#endif

#include "dwc_otg_driver.h"
#include "dwc_otg_cil.h"

#ifndef CONFIG_IFX
#include "dwc_otg_pcd.h"
#include "dwc_otg_hcd.h"
#else
#include "dwc_otg_pcd.h" // device
//#include "host/dwc_otg_hcd.h"   // host
#include "dwc_otg_ifx.h" // for Infineon platform specific.
#endif

#define	DWC_DRIVER_VERSION	"2.40a 10-APR-2006"
#define	DWC_DRIVER_DESC		"HS OTG USB Controller driver"

// 20070316, winder added.
#ifdef CONFIG_IFX
#define SZ_256K                         0x00040000
#endif

static const char dwc_driver_name[] = "dwc_otg";
// 20070316, winder added.
#ifdef CONFIG_IFX
/* Winder: default is for Amazon-SE system.
 * If other platform need to use USB, please pass modules param.
 */
static unsigned long dwc_iomem_base = 0x1e101000;
static int dwc_irq = IFX_USB_IRQ;
static unsigned long dwc_param_opt = 1;
static unsigned long dwc_param_otg_cap = 2;
static unsigned long dwc_param_dma_enable = 1;
static unsigned long dwc_param_dma_burst_size = 32;
static unsigned long dwc_param_speed = 0;
static unsigned long dwc_param_host_support_fs_ls_low_power = 0;
static unsigned long dwc_param_host_ls_low_power_phy_clk = 0;
static unsigned long dwc_param_enable_dynamic_fifo = 1;
static unsigned long dwc_param_data_fifo_size = 512;
#ifdef USE_PERIODIC_EP
static unsigned long dwc_param_dev_rx_fifo_size = 240;
static unsigned long dwc_param_dev_nperio_tx_fifo_size = 240;
static unsigned long dwc_param_dev_perio_tx_fifo_size = 32;
#else
static unsigned long dwc_param_dev_rx_fifo_size = 256;
static unsigned long dwc_param_dev_nperio_tx_fifo_size = 256;
#endif

static unsigned long dwc_param_host_rx_fifo_size = 256;
static unsigned long dwc_param_host_nperio_tx_fifo_size = 128;
static unsigned long dwc_param_host_perio_tx_fifo_size = 128;
static unsigned long dwc_param_max_transfer_size = 65535;
static unsigned long dwc_param_max_packet_count = 511;
static unsigned long dwc_param_host_channels = 16;
static unsigned long dwc_param_dev_endpoints = 6;
static unsigned long dwc_param_phy_type = 1;
static unsigned long dwc_param_phy_utmi_width = 16;
static unsigned long dwc_param_phy_ulpi_ddr = 0;
static unsigned long dwc_param_phy_ulpi_ext_vbus = 0;
static unsigned long dwc_param_i2c_enable = 0;
static unsigned long dwc_param_ulpi_fs_ls = 0;
static unsigned long dwc_param_ts_dline = 0;
#endif // CONFIG_IFX

#ifdef HDEBUG
int HDEBUGCNT=0;
#endif

/*-------------------------------------------------------------------------*/
/* Encapsulate the module parameter settings */

static dwc_otg_core_params_t dwc_otg_module_params = {
        .opt = -1,
        .otg_cap = -1,
        .dma_enable = -1,
	.dma_burst_size = -1,
	.speed = -1,
	.host_support_fs_ls_low_power = -1,
	.host_ls_low_power_phy_clk = -1,
	.enable_dynamic_fifo = -1,
	.data_fifo_size = -1,
	.dev_rx_fifo_size = -1,
	.dev_nperio_tx_fifo_size = -1,
	.dev_perio_tx_fifo_size = {-1,   /* dev_perio_tx_fifo_size_1 */
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1,
				   -1},  /* 15 */
	.host_rx_fifo_size = -1,
	.host_nperio_tx_fifo_size = -1,
	.host_perio_tx_fifo_size = -1,
	.max_transfer_size = -1,
	.max_packet_count = -1,
	.host_channels = -1,
	.dev_endpoints = -1,
	.phy_type = -1,
        .phy_utmi_width = -1,
        .phy_ulpi_ddr = -1,
        .phy_ulpi_ext_vbus = -1,
	.i2c_enable = -1,
	.ulpi_fs_ls = -1,
	.ts_dline = -1,
};
#ifdef CONFIG_IFX
static struct proc_dir_entry * proc_ifx_root = NULL;
/*
 * initialize the proc file system and make a dir named /proc/[name]
 */
static void ifx_proc_init(void)
{
	proc_ifx_root = proc_mkdir(dwc_driver_name, (void *)0);
	if (!proc_ifx_root)
	{
		printk("%s proc initialization failed! \n", dwc_driver_name);
		return;
	}
	printk("%s proc initialization okay! \n", dwc_driver_name);
}
/*
 * proc file system add function for debugging.
 */
int ifx_proc_addproc(char *funcname, read_proc_t *hookfuncr, write_proc_t *hookfuncw)
{
	struct proc_dir_entry *pe;
	if (!proc_ifx_root)
		ifx_proc_init();

	if (hookfuncw == NULL)
	{
		pe = create_proc_read_entry(funcname, 0, proc_ifx_root, hookfuncr, NULL);
		if (!pe)
		{
			printk("ERROR in creating read proc entry (%s)! \n", funcname);
			return -1;
		}
	}else
	{
		pe = create_proc_entry(funcname, S_IRUGO | S_IWUGO, proc_ifx_root);
		if (pe)
		{
			pe->read_proc = hookfuncr;
			pe->write_proc = hookfuncw;
		}else
		{
			printk("ERROR in creating proc entry (%s)! \n", funcname);
			return -1;
		}
	}
	return 0;
}
/*
 * proc file system del function for removing module.
 */
void ifx_proc_delproc(char *funcname)
{
    char pname[30];
    sprintf(pname, "%s/%s", dwc_driver_name, funcname);
    remove_proc_entry(pname, proc_ifx_root);
}

#endif

#ifndef CONFIG_IFX // winder, for 2.6 kernel??
/**
 * This function shows the Driver Version.
 */
static ssize_t version_show(struct device_driver *dev, char *buf)
{
        return snprintf(buf, sizeof(DWC_DRIVER_VERSION)+2,"%s\n",
                        DWC_DRIVER_VERSION);
}
static DRIVER_ATTR(version, S_IRUGO, version_show, NULL);
#else // CONFIG_IFX, for proc filesystem.
static int version_show(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int len = 0;
    len += snprintf(buf+len, sizeof(DWC_DRIVER_VERSION)+2,"%s\n",
                DWC_DRIVER_VERSION);

    return len;
}
#endif // CONFIG_IFX

/**
 * Global Debug Level Mask.
 */
//uint32_t g_dbg_lvl = 0x0; /* OFF */
uint32_t g_dbg_lvl = 0xFF; /* OFF */

#ifndef CONFIG_IFX // winder, for kernel 2.6??
/**
 * This function shows the driver Debug Level.
 */
static ssize_t dbg_level_show(struct device_driver *_drv, char *_buf)
{
        return sprintf(_buf, "0x%0x\n", g_dbg_lvl);
}
/**
 * This function stores the driver Debug Level.
 */
static ssize_t dbg_level_store(struct device_driver *_drv, const char *_buf,
                               size_t _count)
{
	g_dbg_lvl = simple_strtoul(_buf, NULL, 16);
        return _count;
}
static DRIVER_ATTR(debuglevel, S_IRUGO|S_IWUSR, dbg_level_show, dbg_level_store);
#else // CONFIG_IFX, for proc filesystem.
static int dbg_level_show(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int len = 0;

    len += sprintf(buf+len, "0x%0x\n", g_dbg_lvl);

    return len;
}

static int dbg_level_store(struct file *file, const char *buffer, unsigned long count, void *data)
{
    char buf[10];
    int i = 0;

    if (copy_from_user(buf, &buffer[i], sizeof(buf))){
        return -EFAULT;
    }

    g_dbg_lvl = simple_strtoul(buf, NULL, 16);

    return count;
}
#endif // CONFIG_IFX

/**
 * This function is called during module intialization to verify that
 * the module parameters are in a valid state.
 */
static int check_parameters(dwc_otg_core_if_t *core_if)
{
	int i;
	int retval = 0;

/* Checks if the parameter is outside of its valid range of values */
#define DWC_OTG_PARAM_TEST(_param_,_low_,_high_) \
	((dwc_otg_module_params._param_ < (_low_)) || \
         (dwc_otg_module_params._param_ > (_high_)))

/* If the parameter has been set by the user, check that the parameter value is
 * within the value range of values.  If not, report a module error. */
#define DWC_OTG_PARAM_ERR(_param_,_low_,_high_,_string_) \
        do { \
		if (dwc_otg_module_params._param_ != -1) { \
			if (DWC_OTG_PARAM_TEST(_param_,(_low_),(_high_))) { \
				DWC_ERROR("`%d' invalid for parameter `%s'\n", \
					  dwc_otg_module_params._param_, _string_); \
				dwc_otg_module_params._param_ = dwc_param_##_param_##_default; \
				retval ++; \
			} \
		} \
	} while (0)

	DWC_OTG_PARAM_ERR(opt,0,1,"opt");
	DWC_OTG_PARAM_ERR(otg_cap,0,2,"otg_cap");
        DWC_OTG_PARAM_ERR(dma_enable,0,1,"dma_enable");
	DWC_OTG_PARAM_ERR(speed,0,1,"speed");
	DWC_OTG_PARAM_ERR(host_support_fs_ls_low_power,0,1,"host_support_fs_ls_low_power");
	DWC_OTG_PARAM_ERR(host_ls_low_power_phy_clk,0,1,"host_ls_low_power_phy_clk");
	DWC_OTG_PARAM_ERR(enable_dynamic_fifo,0,1,"enable_dynamic_fifo");
	DWC_OTG_PARAM_ERR(data_fifo_size,32,32768,"data_fifo_size");
	DWC_OTG_PARAM_ERR(dev_rx_fifo_size,16,32768,"dev_rx_fifo_size");
	DWC_OTG_PARAM_ERR(dev_nperio_tx_fifo_size,16,32768,"dev_nperio_tx_fifo_size");
	DWC_OTG_PARAM_ERR(host_rx_fifo_size,16,32768,"host_rx_fifo_size");
	DWC_OTG_PARAM_ERR(host_nperio_tx_fifo_size,16,32768,"host_nperio_tx_fifo_size");
	DWC_OTG_PARAM_ERR(host_perio_tx_fifo_size,16,32768,"host_perio_tx_fifo_size");
	DWC_OTG_PARAM_ERR(max_transfer_size,2047,524288,"max_transfer_size");
	DWC_OTG_PARAM_ERR(max_packet_count,15,511,"max_packet_count");
	DWC_OTG_PARAM_ERR(host_channels,1,16,"host_channels");
	DWC_OTG_PARAM_ERR(dev_endpoints,1,15,"dev_endpoints");
	DWC_OTG_PARAM_ERR(phy_type,0,2,"phy_type");
        DWC_OTG_PARAM_ERR(phy_ulpi_ddr,0,1,"phy_ulpi_ddr");
        DWC_OTG_PARAM_ERR(phy_ulpi_ext_vbus,0,1,"phy_ulpi_ext_vbus");
	DWC_OTG_PARAM_ERR(i2c_enable,0,1,"i2c_enable");
	DWC_OTG_PARAM_ERR(ulpi_fs_ls,0,1,"ulpi_fs_ls");
	DWC_OTG_PARAM_ERR(ts_dline,0,1,"ts_dline");

	if (dwc_otg_module_params.dma_burst_size != -1) {
		if (DWC_OTG_PARAM_TEST(dma_burst_size,1,1) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,4,4) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,8,8) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,16,16) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,32,32) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,64,64) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,128,128) &&
		    DWC_OTG_PARAM_TEST(dma_burst_size,256,256))
		{
			DWC_ERROR("`%d' invalid for parameter `dma_burst_size'\n",
				  dwc_otg_module_params.dma_burst_size);
			dwc_otg_module_params.dma_burst_size = 32;
			retval ++;
		}
	}

	if (dwc_otg_module_params.phy_utmi_width != -1) {
		if (DWC_OTG_PARAM_TEST(phy_utmi_width,8,8) &&
		    DWC_OTG_PARAM_TEST(phy_utmi_width,16,16))
		{
			DWC_ERROR("`%d' invalid for parameter `phy_utmi_width'\n",
				  dwc_otg_module_params.phy_utmi_width);
			dwc_otg_module_params.phy_utmi_width = 16;
			retval ++;
		}
	}

#ifdef USE_PERIODIC_EP
	for (i=0; i<15; i++) {
		/** @todo should be like above */
		//DWC_OTG_PARAM_ERR(dev_perio_tx_fifo_size[i],4,768,"dev_perio_tx_fifo_size");
		if (dwc_otg_module_params.dev_perio_tx_fifo_size[i] != -1) {
			if (DWC_OTG_PARAM_TEST(dev_perio_tx_fifo_size[i],4,768)) {
				DWC_ERROR("`%d' invalid for parameter `%s_%d'\n",
					  dwc_otg_module_params.dev_perio_tx_fifo_size[i], "dev_perio_tx_fifo_size", i);
				dwc_otg_module_params.dev_perio_tx_fifo_size[i] = dwc_param_dev_perio_tx_fifo_size_default;
				retval ++;
			}
		}
	}
#endif


	/* At this point, all module parameters that have been set by the user
	 * are valid, and those that have not are left unset.  Now set their
	 * default values and/or check the parameters against the hardware
	 * configurations of the OTG core. */



/* This sets the parameter to the default value if it has not been set by the
 * user */
#define DWC_OTG_PARAM_SET_DEFAULT(_param_) \
	({ \
		int changed = 1; \
		if (dwc_otg_module_params._param_ == -1) { \
			changed = 0; \
			dwc_otg_module_params._param_ = dwc_param_##_param_##_default; \
		} \
		changed; \
	})

/* This checks the macro agains the hardware configuration to see if it is
 * valid.  It is possible that the default value could be invalid.  In this
 * case, it will report a module error if the user touched the parameter.
 * Otherwise it will adjust the value without any error. */
#define DWC_OTG_PARAM_CHECK_VALID(_param_,_str_,_is_valid_,_set_valid_) \
	({ \
	        int changed = DWC_OTG_PARAM_SET_DEFAULT(_param_); \
		int error = 0; \
		if (!(_is_valid_)) { \
			if (changed) { \
				DWC_ERROR("`%d' invalid for parameter `%s'.Check HW configuration.\n", dwc_otg_module_params._param_,_str_); \
				error = 1; \
			} \
			dwc_otg_module_params._param_ = (_set_valid_); \
		} \
		error; \
	})

	/* OTG Cap */
	retval += DWC_OTG_PARAM_CHECK_VALID(otg_cap,"otg_cap",
                  ({
			  int valid;
			  valid = 1;
			  switch (dwc_otg_module_params.otg_cap) {
			  case DWC_OTG_CAP_PARAM_HNP_SRP_CAPABLE:
				  if (core_if->hwcfg2.b.op_mode != DWC_HWCFG2_OP_MODE_HNP_SRP_CAPABLE_OTG) valid = 0;
				  break;
			  case DWC_OTG_CAP_PARAM_SRP_ONLY_CAPABLE:
				  if ((core_if->hwcfg2.b.op_mode != DWC_HWCFG2_OP_MODE_HNP_SRP_CAPABLE_OTG) &&
				      (core_if->hwcfg2.b.op_mode != DWC_HWCFG2_OP_MODE_SRP_ONLY_CAPABLE_OTG) &&
				      (core_if->hwcfg2.b.op_mode != DWC_HWCFG2_OP_MODE_SRP_CAPABLE_DEVICE) &&
				      (core_if->hwcfg2.b.op_mode != DWC_HWCFG2_OP_MODE_SRP_CAPABLE_HOST))
				  {
					  valid = 0;
				  }
				  break;
			  case DWC_OTG_CAP_PARAM_NO_HNP_SRP_CAPABLE:
				  /* always valid */
				  break;
			  }
			  valid;
		  }),
                  (((core_if->hwcfg2.b.op_mode == DWC_HWCFG2_OP_MODE_HNP_SRP_CAPABLE_OTG) ||
		    (core_if->hwcfg2.b.op_mode == DWC_HWCFG2_OP_MODE_SRP_ONLY_CAPABLE_OTG) ||
		    (core_if->hwcfg2.b.op_mode == DWC_HWCFG2_OP_MODE_SRP_CAPABLE_DEVICE) ||
		    (core_if->hwcfg2.b.op_mode == DWC_HWCFG2_OP_MODE_SRP_CAPABLE_HOST)) ?
		   DWC_OTG_CAP_PARAM_SRP_ONLY_CAPABLE :
		   DWC_OTG_CAP_PARAM_NO_HNP_SRP_CAPABLE));

	retval += DWC_OTG_PARAM_CHECK_VALID(dma_enable,"dma_enable",
					    ((dwc_otg_module_params.dma_enable == 1) && (core_if->hwcfg2.b.architecture == 0)) ? 0 : 1,
					    0);

	retval += DWC_OTG_PARAM_CHECK_VALID(opt,"opt",
					    1,
					    0);

	DWC_OTG_PARAM_SET_DEFAULT(dma_burst_size);

	retval += DWC_OTG_PARAM_CHECK_VALID(host_support_fs_ls_low_power,
					    "host_support_fs_ls_low_power",
					    1, 0);

	retval += DWC_OTG_PARAM_CHECK_VALID(enable_dynamic_fifo,
				  "enable_dynamic_fifo",
				  ((dwc_otg_module_params.enable_dynamic_fifo == 0) ||
				   (core_if->hwcfg2.b.dynamic_fifo == 1)), 0);


	retval += DWC_OTG_PARAM_CHECK_VALID(data_fifo_size,
				  "data_fifo_size",
				  (dwc_otg_module_params.data_fifo_size <= core_if->hwcfg3.b.dfifo_depth),
				  core_if->hwcfg3.b.dfifo_depth);

	retval += DWC_OTG_PARAM_CHECK_VALID(dev_rx_fifo_size,
				  "dev_rx_fifo_size",
				  (dwc_otg_module_params.dev_rx_fifo_size <= dwc_read_reg32(&core_if->core_global_regs->grxfsiz)),
				  dwc_read_reg32(&core_if->core_global_regs->grxfsiz));

	retval += DWC_OTG_PARAM_CHECK_VALID(dev_nperio_tx_fifo_size,
				  "dev_nperio_tx_fifo_size",
				  (dwc_otg_module_params.dev_nperio_tx_fifo_size <= (dwc_read_reg32(&core_if->core_global_regs->gnptxfsiz) >> 16)),
				  (dwc_read_reg32(&core_if->core_global_regs->gnptxfsiz) >> 16));

	retval += DWC_OTG_PARAM_CHECK_VALID(host_rx_fifo_size,
					    "host_rx_fifo_size",
					    (dwc_otg_module_params.host_rx_fifo_size <= dwc_read_reg32(&core_if->core_global_regs->grxfsiz)),
					    dwc_read_reg32(&core_if->core_global_regs->grxfsiz));


	retval += DWC_OTG_PARAM_CHECK_VALID(host_nperio_tx_fifo_size,
				  "host_nperio_tx_fifo_size",
				  (dwc_otg_module_params.host_nperio_tx_fifo_size <= (dwc_read_reg32(&core_if->core_global_regs->gnptxfsiz) >> 16)),
				  (dwc_read_reg32(&core_if->core_global_regs->gnptxfsiz) >> 16));

	retval += DWC_OTG_PARAM_CHECK_VALID(host_perio_tx_fifo_size,
					    "host_perio_tx_fifo_size",
					    (dwc_otg_module_params.host_perio_tx_fifo_size <= ((dwc_read_reg32(&core_if->core_global_regs->hptxfsiz) >> 16))),
					    ((dwc_read_reg32(&core_if->core_global_regs->hptxfsiz) >> 16)));

	retval += DWC_OTG_PARAM_CHECK_VALID(max_transfer_size,
				  "max_transfer_size",
				  (dwc_otg_module_params.max_transfer_size < (1 << (core_if->hwcfg3.b.xfer_size_cntr_width + 11))),
				  ((1 << (core_if->hwcfg3.b.xfer_size_cntr_width + 11)) - 1));

	retval += DWC_OTG_PARAM_CHECK_VALID(max_packet_count,
				  "max_packet_count",
				  (dwc_otg_module_params.max_packet_count < (1 << (core_if->hwcfg3.b.packet_size_cntr_width + 4))),
				  ((1 << (core_if->hwcfg3.b.packet_size_cntr_width + 4)) - 1));

	retval += DWC_OTG_PARAM_CHECK_VALID(host_channels,
				  "host_channels",
				  (dwc_otg_module_params.host_channels <= (core_if->hwcfg2.b.num_host_chan + 1)),
				  (core_if->hwcfg2.b.num_host_chan + 1));

	retval += DWC_OTG_PARAM_CHECK_VALID(dev_endpoints,
				  "dev_endpoints",
				  (dwc_otg_module_params.dev_endpoints <= (core_if->hwcfg2.b.num_dev_ep)),
				  core_if->hwcfg2.b.num_dev_ep);

/*
 * Define the following to disable the FS PHY Hardware checking.  This is for
 * internal testing only.
 *
 * #define NO_FS_PHY_HW_CHECKS
 */

#ifdef NO_FS_PHY_HW_CHECKS
	retval += DWC_OTG_PARAM_CHECK_VALID(phy_type,
					    "phy_type", 1, 0);
#else
	retval += DWC_OTG_PARAM_CHECK_VALID(phy_type,
					    "phy_type",
					    ({
						    int valid = 0;
						    if ((dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_UTMI) &&
							((core_if->hwcfg2.b.hs_phy_type == 1) ||
							 (core_if->hwcfg2.b.hs_phy_type == 3)))
						    {
							    valid = 1;
						    }
						    else if ((dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_ULPI) &&
							     ((core_if->hwcfg2.b.hs_phy_type == 2) ||
							      (core_if->hwcfg2.b.hs_phy_type == 3)))
						    {
							    valid = 1;
						    }
						    else if ((dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_FS) &&
							     (core_if->hwcfg2.b.fs_phy_type == 1))
						    {
							    valid = 1;
						    }
						    valid;
					    }),
					    ({
						    int set = DWC_PHY_TYPE_PARAM_FS;
						    if (core_if->hwcfg2.b.hs_phy_type) {
							    if ((core_if->hwcfg2.b.hs_phy_type == 3) ||
								(core_if->hwcfg2.b.hs_phy_type == 1)) {
								    set = DWC_PHY_TYPE_PARAM_UTMI;
							    }
							    else {
								    set = DWC_PHY_TYPE_PARAM_ULPI;
							    }
						    }
						    set;
					    }));
#endif

	retval += DWC_OTG_PARAM_CHECK_VALID(speed,"speed",
					    (dwc_otg_module_params.speed == 0) && (dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_FS) ? 0 : 1,
					    dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_FS ? 1 : 0);

	retval += DWC_OTG_PARAM_CHECK_VALID(host_ls_low_power_phy_clk,
					    "host_ls_low_power_phy_clk",
					    ((dwc_otg_module_params.host_ls_low_power_phy_clk == DWC_HOST_LS_LOW_POWER_PHY_CLK_PARAM_48MHZ) && (dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_FS) ? 0 : 1),
					    ((dwc_otg_module_params.phy_type == DWC_PHY_TYPE_PARAM_FS) ? DWC_HOST_LS_LOW_POWER_PHY_CLK_PARAM_6MHZ : DWC_HOST_LS_LOW_POWER_PHY_CLK_PARAM_48MHZ));

        DWC_OTG_PARAM_SET_DEFAULT(phy_ulpi_ddr);
        DWC_OTG_PARAM_SET_DEFAULT(phy_ulpi_ext_vbus);
        DWC_OTG_PARAM_SET_DEFAULT(phy_utmi_width);
        DWC_OTG_PARAM_SET_DEFAULT(ulpi_fs_ls);
        DWC_OTG_PARAM_SET_DEFAULT(ts_dline);

#ifdef NO_FS_PHY_HW_CHECKS
	retval += DWC_OTG_PARAM_CHECK_VALID(i2c_enable,
					    "i2c_enable", 1, 0);
#else
	retval += DWC_OTG_PARAM_CHECK_VALID(i2c_enable,
					    "i2c_enable",
					    (dwc_otg_module_params.i2c_enable == 1) && (core_if->hwcfg3.b.i2c == 0) ? 0 : 1,
					    0);
#endif

#ifdef USE_PERIODIC_EP
	for (i=0; i<16; i++) {
		int changed = 1;
		int error = 0;

		if (dwc_otg_module_params.dev_perio_tx_fifo_size[i] == -1) {
			changed = 0;
			dwc_otg_module_params.dev_perio_tx_fifo_size[i] = dwc_param_dev_perio_tx_fifo_size_default;
		}
		if (!(dwc_otg_module_params.dev_perio_tx_fifo_size[i] <= (dwc_read_reg32(&core_if->core_global_regs->dptxfsiz[i])))) {
			if (changed) {
				DWC_ERROR("`%d' invalid for parameter `dev_perio_fifo_size_%d'.  Check HW configuration.\n", dwc_otg_module_params.dev_perio_tx_fifo_size[i],i);
				error = 1;
			}
			dwc_otg_module_params.dev_perio_tx_fifo_size[i] = dwc_read_reg32(&core_if->core_global_regs->dptxfsiz[i]);
		}
		retval += error;
	}
#endif

	return retval;
} // check_parameters


/**
 * This function is the top level interrupt handler for the Common
 * (Device and host modes) interrupts.
 */
static irqreturn_t dwc_otg_common_irq(int _irq, void *_dev, struct pt_regs *_r)
{
	dwc_otg_device_t *otg_dev = _dev;
#ifdef CONFIG_IFX // winder.
	int32_t retval;
#else
	int32_t retval = IRQ_NONE;
#endif

	retval = dwc_otg_handle_common_intr( otg_dev->core_if );

#ifdef CONFIG_IFX // winder.
	mask_and_ack_ifx_irq (_irq);
#endif
	return IRQ_RETVAL(retval);
}

#ifdef CONFIG_IFX // winder
static unsigned long pri_drvdata = 0;

/**
 * This function is called when a DWC_OTG device is registerred and need to get private
 * device data.
 *
 * @return (dwc_otg_device_t *) pri_drvdata.
 */
dwc_otg_device_t *dwc_get_drvdata(void) // winder, get device data.
{
HDEBUG_SE(0);
    return (dwc_otg_device_t *)pri_drvdata;
}

/**
 * This function is called when a DWC_OTG device is registerred and need to set private
 * device data.
 *
 * @return
 */
void dwc_set_drvdata(dwc_otg_device_t *drvdata) // winder, set device data.
{
HDEBUG_SE(0);
    pri_drvdata = (unsigned long)drvdata;
}

/**
 * This function is called when a DWC_OTG device is registerred and need to get module
 * parameter data.
 *
 * @return (dwc_otg_core_params_t *) dwc_otg_module_params.
 */
dwc_otg_core_params_t *dwc_otg_get_module_params(void)
{
HDEBUG_SE(0);
    return (dwc_otg_core_params_t *) &dwc_otg_module_params;
}

#endif

/**
 * This function is called when a DWC_OTG device is unregistered with the
 * dwc_otg_driver. This happens, for example, when the rmmod command is
 * executed. The device may or may not be electrically present. If it is
 * present, the driver stops device processing. Any resources used on behalf
 * of this device are freed.
 *
 * @return
 */
#ifndef CONFIG_IFX // winder
static void dwc_otg_driver_remove(struct lm_device *_lmdev)
{
    dwc_otg_device_t *otg_dev = lm_get_drvdata(_lmdev);
HDEBUG_S;
    DWC_DEBUGPL(DBG_ANY, "%s(%p)\n", __func__, _lmdev);

    if (otg_dev == NULL) {
        /* Memory allocation for the dwc_otg_device failed. */
HDEBUG_E(1);
        return;
    }

    /*
    * Free the IRQ
    */
    if (otg_dev->common_irq_installed) {
        free_irq( _lmdev->irq, otg_dev );
    }

#ifndef DWC_DEVICE_ONLY
    if (otg_dev->hcd != NULL) {
        dwc_otg_hcd_remove( _lmdev );
    }
#endif

#ifndef DWC_HOST_ONLY
    if (otg_dev->pcd != NULL) {
        dwc_otg_pcd_remove( _lmdev );
    }
#endif
    if (otg_dev->core_if != NULL) {
        dwc_otg_cil_remove( otg_dev->core_if );
    }

    /*
     * Remove the device attributes
     */
    dwc_otg_attr_remove(_lmdev);

    /*
     * Return the memory.
     */
    if (otg_dev->base != NULL) {
        iounmap(otg_dev->base);
    }
    kfree(otg_dev);

    /*
     * Clear the drvdata pointer.
     */
    lm_set_drvdata( _lmdev, 0 );
HDEBUG_E(0);
}

#else // CONFIG_IFX
static void dwc_otg_driver_remove(void)
{
    dwc_otg_device_t *otg_dev = lm_get_drvdata(_lmdev);

HDEBUG_S;
    DWC_DEBUGPL(DBG_ANY, "%s(%p)\n", __func__, otg_dev);
    if (otg_dev == NULL) {
        /* Memory allocation for the dwc_otg_device failed. */
HDEBUG_E(1);
        return;
    }

    /*
    * Free the IRQ
    */
    if (otg_dev->common_irq_installed) {
        free_irq( otg_dev->irq, otg_dev );
    }
#ifndef DWC_DEVICE_ONLY
    if (otg_dev->hcd != NULL) {
        dwc_otg_hcd_remove();
    }
#endif

#ifndef DWC_HOST_ONLY
    if (otg_dev->pcd != NULL) {
        dwc_otg_pcd_remove(otg_dev);
    }
#endif

    /*
     * Return the memory.
     */
    if (otg_dev->base != NULL) {
        iounmap(otg_dev->base);
    }
    kfree(otg_dev);
HDEBUG_E(0);
}
#endif // CONFIG_IFX
/**
 * This function is called when an DWC_OTG device is bound to a
 * dwc_otg_driver. It creates the driver components required to
 * control the device (CIL, HCD, and PCD) and it initializes the
 * device. The driver components are stored in a dwc_otg_device
 * structure. A reference to the dwc_otg_device is saved in the
 * lm_device. This allows the driver to access the dwc_otg_device
 * structure on subsequent calls to driver methods for this device.
 *
 * @return
 */
#ifndef CONFIG_IFX // winder
static int dwc_otg_driver_probe(struct lm_device *_lmdev)
#else
static int dwc_otg_driver_probe(void)
#endif
{
    int retval = 0;
    dwc_otg_device_t *dwc_otg_device;
    int32_t	snpsid;

HDEBUG_S;


typedef union endiantest_data
{
        /** raw register data */
        uint32_t d32;
        /** register bits */
        struct
        {
		unsigned bit31 : 1;
		unsigned bit30 : 1;
		unsigned bit29 : 1;
		unsigned bit28 : 1;
		unsigned bit27 : 1;
		unsigned bit26 : 1;
		unsigned bit25 : 1;
		unsigned bit24 : 1;
		unsigned bit23 : 1;
		unsigned bit22 : 1;
		unsigned bit21 : 1;
		unsigned bit20 : 1;
		unsigned bit19 : 1;
		unsigned bit18 : 1;
		unsigned bit17 : 1;
		unsigned bit16 : 1;
		unsigned bit15 : 1;
		unsigned bit14 : 1;
		unsigned bit13 : 1;
		unsigned bit12 : 1;
		unsigned bit11 : 1;
		unsigned bit10 : 1;
		unsigned bit09 : 1;
		unsigned bit08 : 1;
		unsigned bit07 : 1;
		unsigned bit06 : 1;
		unsigned bit05 : 1;
		unsigned bit04 : 1;
		unsigned bit03 : 1;
		unsigned bit02 : 1;
		unsigned bit01 : 1;
		unsigned bit00 : 1;
        } b[2];
} endiantest_data_t;






#ifndef CONFIG_IFX // winder
    dev_dbg(&_lmdev->dev, "dwc_otg_driver_probe(%p)\n", _lmdev);
    dev_dbg(&_lmdev->dev, "start=0x%08x\n", (unsigned)_lmdev->resource.start);
#else
    dbg("dwc_otg_driver_probe\n");
    dev_dbg(&_lmdev->dev, "start=0x%08x\n", (unsigned int)dwc_iomem_base);
#endif

    dwc_otg_device = kmalloc(sizeof(dwc_otg_device_t), GFP_KERNEL);
    if (dwc_otg_device == 0) {
        dev_err(&_lmdev->dev, "kmalloc of dwc_otg_device failed\n");
        retval = -ENOMEM;
        goto fail;
    }
    memset(dwc_otg_device, 0, sizeof(*dwc_otg_device));
    dwc_otg_device->reg_offset = 0xFFFFFFFF;

    /*
     * Map the DWC_otg Core memory into virtual address space.
     */
#ifndef CONFIG_IFX // winder
    dwc_otg_device->base = ioremap(_lmdev->resource.start, SZ_256K);
#else
    dwc_otg_device->base = ioremap_nocache(dwc_iomem_base, SZ_256K);
    dwc_otg_device->reg_size = SZ_256K;
#endif
    if (dwc_otg_device->base == NULL)
    {
        dev_err(&_lmdev->dev, "ioremap() failed\n");
        retval = -ENOMEM;
        goto fail;
    }
    dev_dbg(&_lmdev->dev, "base=0x%08x\n", (unsigned)dwc_otg_device->base);


    /*
     * Attempt to ensure this device is really a DWC_otg Controller.
     * Read and verify the SNPSID register contents. The value should be
     * 0x45F42XXX, which corresponds to "OT2", as in "OTG version 2.XX".
     */
    snpsid = dwc_read_reg32((uint32_t *)((uint8_t *)dwc_otg_device->base + 0x40));
    if ((snpsid & 0xFFFFF000) != 0x4F542000) {
        dev_err(&_lmdev->dev, "Bad value for SNPSID: 0x%08x\n", snpsid);
        retval = -EINVAL;
        goto fail;
    }

    /*
     * Initialize driver data to point to the global DWC_otg
     * Device structure.
     */
#ifndef CONFIG_IFX // winder
    lm_set_drvdata( _lmdev, dwc_otg_device );
#endif
    dev_dbg(&_lmdev->dev, "dwc_otg_device=0x%p\n", dwc_otg_device);

    dwc_otg_device->core_if = dwc_otg_cil_init( dwc_otg_device->base,
        &dwc_otg_module_params);
    if (dwc_otg_device->core_if == 0) {
        dev_err(&_lmdev->dev, "CIL initialization failed!\n");
        retval = -ENOMEM;
        goto fail;
    }

    /*
     * Validate parameter values.
     */
    if (check_parameters(dwc_otg_device->core_if) != 0) {
        retval = -EINVAL;
        goto fail;
    }

#ifndef CONFIG_IFX // winder.
    /*
     * Create Device Attributes in sysfs
     */
    dwc_otg_attr_create (_lmdev);
#endif

    /*
     * Disable the global interrupt until all the interrupt
     * handlers are installed.
     */
    dwc_otg_disable_global_interrupts( dwc_otg_device->core_if );
    /*
     * Install the interrupt handler for the common interrupts before
     * enabling common interrupts in core_init below.
     */
#ifndef CONFIG_IFX // winder.
    DWC_DEBUGPL( DBG_CIL, "registering (common) handler for irq%d\n",
        _lmdev->irq);
    retval = request_irq(_lmdev->irq, dwc_otg_common_irq,
        SA_SHIRQ, "dwc_otg", dwc_otg_device );
    if (retval != 0) {
        DWC_ERROR("request of irq%d failed\n", _lmdev->irq);
        retval = -EBUSY;
        goto fail;
    } else {
        dwc_otg_device->common_irq_installed = 1;
    }
#else // CONFIG_IFX
    DWC_DEBUGPL( DBG_CIL, "registering (common) handler for irq%d\n",
        dwc_irq);
    retval = request_irq(dwc_irq, dwc_otg_common_irq,
        SA_SHIRQ, "dwc_otg", dwc_otg_device );
    if (retval != 0) {
        DWC_ERROR("request of irq%d failed\n", dwc_irq);
        retval = -EBUSY;
        goto fail;
    } else {
        dwc_otg_device->common_irq_installed = 1;
        dwc_otg_device->irq = dwc_irq;
    }
#endif // CONFIG_IFX

#ifdef CONFIG_MACH_IPMATE
    set_irq_type(_lmdev->irq, IRQT_LOW);
#endif

    /*
     * Initialize the DWC_otg core.
     */
    dwc_otg_core_init( dwc_otg_device->core_if );

#ifndef DWC_HOST_ONLY  // otg device mode. (gadget.)
    /*
     * Initialize the PCD
     */
#ifndef CONFIG_IFX // winder.
    retval = dwc_otg_pcd_init( _lmdev );
#else
    dwc_set_drvdata(dwc_otg_device);
    retval = dwc_otg_pcd_init(dwc_otg_device);
#endif
    if (retval != 0) {
        DWC_ERROR("dwc_otg_pcd_init failed\n");
        dwc_otg_device->pcd = NULL;
        goto fail;
    }
#endif // DWC_HOST_ONLY
#ifndef DWC_DEVICE_ONLY // otg host mode. (HCD)
    /*
     * Initialize the HCD
     */
#ifndef CONFIG_IFX // winder.
    retval = dwc_otg_hcd_init(_lmdev);
#else
    dwc_set_drvdata(dwc_otg_device);
    retval = dwc_otg_hcd_init(dwc_otg_device);
#endif
    if (retval != 0) {
        DWC_ERROR("dwc_otg_hcd_init failed\n");
        dwc_otg_device->hcd = NULL;
        goto fail;
    }
#endif // DWC_DEVICE_ONLY

    /*
     * Enable the global interrupt after all the interrupt
     * handlers are installed.
     */

#ifndef DWC_DEVICE_ONLY // howard, late enable test
    dwc_otg_enable_global_interrupts( dwc_otg_device->core_if );
#endif

HDEBUG_E(1);
    return 0;

fail:
#ifndef CONFIG_IFX // winder
    dwc_otg_driver_remove(_lmdev);
#else
    dwc_otg_driver_remove();
#endif
HDEBUG_E(2);
    return retval;
}

/**
 * This structure defines the methods to be called by a bus driver
 * during the lifecycle of a device on that bus. Both drivers and
 * devices are registered with a bus driver. The bus driver matches
 * devices to drivers based on information in the device and driver
 * structures.
 *
 * The probe function is called when the bus driver matches a device
 * to this driver. The remove function is called when a device is
 * unregistered with the bus driver.
 */
#ifndef CONFIG_IFX // 20070316, winder, remove lm_device because we are MIPS not ARM.
static struct lm_driver dwc_otg_driver = {
	.drv = {
		.name	= (char*)dwc_driver_name,
	},
	.probe		= dwc_otg_driver_probe,
	.remove		= dwc_otg_driver_remove,
};
#endif

#if defined(CONFIG_IFX) && defined(USE_PROC_FS)
extern void dwc_otg_proc_init(void);
extern void dwc_otg_proc_clean(void);
#endif
/**
 * This function is called when the dwc_otg_driver is installed with the
 * insmod command. It registers the dwc_otg_driver structure with the
 * appropriate bus driver. This will cause the dwc_otg_driver_probe function
 * to be called. In addition, the bus driver will automatically expose
 * attributes defined for the device and driver in the special sysfs file
 * system.
 *
 * @return
 */
static int __init dwc_otg_driver_init(void)
{
    int retval = 0;

HDEBUG_S;
    printk(KERN_INFO "%s: version %s\n", dwc_driver_name, DWC_DRIVER_VERSION);
#ifndef CONFIG_IFX	// 20070316, winder, it is for 2.6 kernel.
    retval = lm_driver_register(&dwc_otg_driver);
#else
    dwc_otg_power_on ();
    retval = dwc_otg_driver_probe();

#if defined(USE_PROC_FS)
    // add proc filesystem.
    ifx_proc_addproc("version", version_show, NULL);
    ifx_proc_addproc("dbg_level", dbg_level_show, dbg_level_store);
    // Winder
    // except version_show and dbg_level, I put proc init to dwc_otg_proc.c.
    dwc_otg_proc_init();
#endif

#endif // CONFIG_IFX
    if (retval < 0) {
        printk(KERN_ERR "%s retval=%d\n", __func__, retval);
HDEBUG_E(1);
        return retval;
    }
#ifndef CONFIG_IFX	// 20070316, winder, it is for 2.6 kernel.
    driver_create_file(&dwc_otg_driver.drv, &driver_attr_version);
    driver_create_file(&dwc_otg_driver.drv, &driver_attr_debuglevel);
#endif

HDEBUG_E(2);
    return retval;
}
module_init(dwc_otg_driver_init);

/**
 * This function is called when the driver is removed from the kernel
 * with the rmmod command. The driver unregisters itself with its bus
 * driver.
 *
 */
static void __exit dwc_otg_driver_cleanup(void)
{
HDEBUG_S;
    printk(KERN_DEBUG "dwc_otg_driver_cleanup()\n");

#ifndef CONFIG_IFX	// 20070316, winder, it is for 2.6 kernel.
    driver_remove_file(&dwc_otg_driver.drv, &driver_attr_debuglevel);
    driver_remove_file(&dwc_otg_driver.drv, &driver_attr_version);

    lm_driver_unregister(&dwc_otg_driver);
#else
    // remove proc filesystem.
#if defined(USE_PROC_FS)
    ifx_proc_delproc("version");
    ifx_proc_delproc("dbg_level");
    dwc_otg_proc_clean();
    remove_proc_entry(dwc_driver_name, 0);
#endif // USE_PROC_FS
    dwc_otg_driver_remove();
#endif

    printk(KERN_INFO "%s module removed\n", dwc_driver_name);
HDEBUG_E(0);

}
module_exit(dwc_otg_driver_cleanup);

MODULE_DESCRIPTION(DWC_DRIVER_DESC);
MODULE_AUTHOR("Synopsys Inc.");
MODULE_LICENSE("GPL");

#ifdef CONFIG_IFX // 20070316, winder removed.
/* TODO: make all parameters configurable from user. */
MODULE_PARM (dwc_iomem_base, "l");
MODULE_PARM_DESC (dwc_iomem_base, "The base address of the DWC_OTG register.");
MODULE_PARM (dwc_irq, "i");
MODULE_PARM_DESC (dwc_irq, "The interrupt number");
MODULE_PARM(g_dbg_lvl, "i");
MODULE_PARM_DESC(g_dbg_lvl, "Debug level.");
MODULE_PARM (dwc_param_opt, "l");
MODULE_PARM_DESC (dwc_param_opt, "OPT Mode");
MODULE_PARM (dwc_param_otg_cap, "l");
MODULE_PARM_DESC (dwc_param_otg_cap, "OTG Capabilities 0=HNP&SRP 1=SRP Only 2=None");
MODULE_PARM(dwc_param_dma_enable, "l");
MODULE_PARM_DESC(dwc_param_dma_enable, "DMA Mode 0=Slave 1=DMA enabled");
MODULE_PARM(dwc_param_dma_burst_size, "l");
MODULE_PARM_DESC(dwc_param_dma_burst_size, "DMA Burst Size 1, 4, 8, 16, 32, 64, 128, 256");
MODULE_PARM(dwc_param_speed, "l");
MODULE_PARM_DESC(dwc_param_speed, "Speed 0=High Speed 1=Full Speed");
MODULE_PARM(dwc_param_host_support_fs_ls_low_power, "l");
MODULE_PARM_DESC(dwc_param_host_support_fs_ls_low_power, "Support Low Power w/FS or LS 0=Support 1=Don't Support");
MODULE_PARM(dwc_param_host_ls_low_power_phy_clk, "l");
MODULE_PARM_DESC(dwc_param_host_ls_low_power_phy_clk, "Low Speed Low Power Clock 0=48Mhz 1=6Mhz");
MODULE_PARM(dwc_param_enable_dynamic_fifo, "l");
MODULE_PARM_DESC(dwc_param_enable_dynamic_fifo, "0=cC Setting 1=Allow Dynamic Sizing");
MODULE_PARM(dwc_param_data_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_data_fifo_size, "Total number of words in the data FIFO memory 32-32768");
MODULE_PARM(dwc_param_dev_rx_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_dev_rx_fifo_size, "Number of words in the Rx FIFO 16-32768");
MODULE_PARM(dwc_param_dev_nperio_tx_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_dev_nperio_tx_fifo_size, "Number of words in the non-periodic Tx FIFO 16-32768");
MODULE_PARM(dwc_param_dev_perio_tx_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_dev_perio_tx_fifo_size, "Number of words in the periodic Tx FIFO 4-768");
MODULE_PARM(dwc_param_host_rx_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_host_rx_fifo_size, "Number of words in the Rx FIFO 16-32768");
MODULE_PARM(dwc_param_host_nperio_tx_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_host_nperio_tx_fifo_size, "Number of words in the non-periodic Tx FIFO 16-32768");
MODULE_PARM(dwc_param_host_perio_tx_fifo_size, "l");
MODULE_PARM_DESC(dwc_param_host_perio_tx_fifo_size, "Number of words in the host periodic Tx FIFO 16-32768");
MODULE_PARM(dwc_param_max_transfer_size, "l");
/** @todo Set the max to 512K, modify checks */
MODULE_PARM_DESC(dwc_param_max_transfer_size, "The maximum transfer size supported in bytes 2047-65535");
MODULE_PARM(dwc_param_max_packet_count, "l");
MODULE_PARM_DESC(dwc_param_max_packet_count, "The maximum number of packets in a transfer 15-511");
MODULE_PARM(dwc_param_host_channels, "l");
MODULE_PARM_DESC(dwc_param_host_channels, "The number of host channel registers to use 1-16");
MODULE_PARM(dwc_param_dev_endpoints, "l");
MODULE_PARM_DESC(dwc_param_dev_endpoints, "The number of endpoints in addition to EP0 available for device mode 1-15");
MODULE_PARM(dwc_param_phy_type, "l");
MODULE_PARM_DESC(dwc_param_phy_type, "0=Reserved 1=UTMI+ 2=ULPI");
MODULE_PARM(dwc_param_phy_utmi_width, "l");
MODULE_PARM_DESC(dwc_param_phy_utmi_width, "Specifies the UTMI+ Data Width 8 or 16 bits");
MODULE_PARM(dwc_param_phy_ulpi_ddr, "l");
MODULE_PARM_DESC(dwc_param_phy_ulpi_ddr, "ULPI at double or single data rate 0=Single 1=Double");
MODULE_PARM(dwc_param_phy_ulpi_ext_vbus, "l");
MODULE_PARM_DESC(dwc_param_phy_ulpi_ext_vbus, "ULPI PHY using internal or external vbus 0=Internal");
MODULE_PARM(dwc_param_i2c_enable, "l");
MODULE_PARM_DESC(dwc_param_i2c_enable, "FS PHY Interface");
MODULE_PARM(dwc_param_ulpi_fs_ls, "l");
MODULE_PARM_DESC(dwc_param_ulpi_fs_ls, "ULPI PHY FS/LS mode only");
MODULE_PARM(dwc_param_ts_dline, "l");
MODULE_PARM_DESC(dwc_param_ts_dline, "Term select Dline pulsing for all PHYs");
#else // CONFIG_IFX
module_param_named(otg_cap, dwc_otg_module_params.otg_cap, int, 0444);
MODULE_PARM_DESC(otg_cap, "OTG Capabilities 0=HNP&SRP 1=SRP Only 2=None");
module_param_named(opt, dwc_otg_module_params.opt, int, 0444);
MODULE_PARM_DESC(opt, "OPT Mode");
module_param_named(dma_enable, dwc_otg_module_params.dma_enable, int, 0444);
MODULE_PARM_DESC(dma_enable, "DMA Mode 0=Slave 1=DMA enabled");
module_param_named(dma_burst_size, dwc_otg_module_params.dma_burst_size, int, 0444);
MODULE_PARM_DESC(dma_burst_size, "DMA Burst Size 1, 4, 8, 16, 32, 64, 128, 256");
module_param_named(speed, dwc_otg_module_params.speed, int, 0444);
MODULE_PARM_DESC(speed, "Speed 0=High Speed 1=Full Speed");
module_param_named(host_support_fs_ls_low_power, dwc_otg_module_params.host_support_fs_ls_low_power, int, 0444);
MODULE_PARM_DESC(host_support_fs_ls_low_power, "Support Low Power w/FS or LS 0=Support 1=Don't Support");
module_param_named(host_ls_low_power_phy_clk, dwc_otg_module_params.host_ls_low_power_phy_clk, int, 0444);
MODULE_PARM_DESC(host_ls_low_power_phy_clk, "Low Speed Low Power Clock 0=48Mhz 1=6Mhz");
module_param_named(enable_dynamic_fifo, dwc_otg_module_params.enable_dynamic_fifo, int, 0444);
MODULE_PARM_DESC(enable_dynamic_fifo, "0=cC Setting 1=Allow Dynamic Sizing");
module_param_named(data_fifo_size, dwc_otg_module_params.data_fifo_size, int, 0444);
MODULE_PARM_DESC(data_fifo_size, "Total number of words in the data FIFO memory 32-32768");
module_param_named(dev_rx_fifo_size, dwc_otg_module_params.dev_rx_fifo_size, int, 0444);
MODULE_PARM_DESC(dev_rx_fifo_size, "Number of words in the Rx FIFO 16-32768");
module_param_named(dev_nperio_tx_fifo_size, dwc_otg_module_params.dev_nperio_tx_fifo_size, int, 0444);
MODULE_PARM_DESC(dev_nperio_tx_fifo_size, "Number of words in the non-periodic Tx FIFO 16-32768");
module_param_named(dev_perio_tx_fifo_size_1, dwc_otg_module_params.dev_perio_tx_fifo_size[0], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_1, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_2, dwc_otg_module_params.dev_perio_tx_fifo_size[1], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_2, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_3, dwc_otg_module_params.dev_perio_tx_fifo_size[2], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_3, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_4, dwc_otg_module_params.dev_perio_tx_fifo_size[3], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_4, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_5, dwc_otg_module_params.dev_perio_tx_fifo_size[4], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_5, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_6, dwc_otg_module_params.dev_perio_tx_fifo_size[5], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_6, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_7, dwc_otg_module_params.dev_perio_tx_fifo_size[6], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_7, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_8, dwc_otg_module_params.dev_perio_tx_fifo_size[7], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_8, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_9, dwc_otg_module_params.dev_perio_tx_fifo_size[8], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_9, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_10, dwc_otg_module_params.dev_perio_tx_fifo_size[9], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_10, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_11, dwc_otg_module_params.dev_perio_tx_fifo_size[10], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_11, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_12, dwc_otg_module_params.dev_perio_tx_fifo_size[11], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_12, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_13, dwc_otg_module_params.dev_perio_tx_fifo_size[12], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_13, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_14, dwc_otg_module_params.dev_perio_tx_fifo_size[13], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_14, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(dev_perio_tx_fifo_size_15, dwc_otg_module_params.dev_perio_tx_fifo_size[14], int, 0444);
MODULE_PARM_DESC(dev_perio_tx_fifo_size_15, "Number of words in the periodic Tx FIFO 4-768");
module_param_named(host_rx_fifo_size, dwc_otg_module_params.host_rx_fifo_size, int, 0444);
MODULE_PARM_DESC(host_rx_fifo_size, "Number of words in the Rx FIFO 16-32768");
module_param_named(host_nperio_tx_fifo_size, dwc_otg_module_params.host_nperio_tx_fifo_size, int, 0444);
MODULE_PARM_DESC(host_nperio_tx_fifo_size, "Number of words in the non-periodic Tx FIFO 16-32768");
module_param_named(host_perio_tx_fifo_size, dwc_otg_module_params.host_perio_tx_fifo_size, int, 0444);
MODULE_PARM_DESC(host_perio_tx_fifo_size, "Number of words in the host periodic Tx FIFO 16-32768");
module_param_named(max_transfer_size, dwc_otg_module_params.max_transfer_size, int, 0444);
/** @todo Set the max to 512K, modify checks */
MODULE_PARM_DESC(max_transfer_size, "The maximum transfer size supported in bytes 2047-65535");
module_param_named(max_packet_count, dwc_otg_module_params.max_packet_count, int, 0444);
MODULE_PARM_DESC(max_packet_count, "The maximum number of packets in a transfer 15-511");
module_param_named(host_channels, dwc_otg_module_params.host_channels, int, 0444);
MODULE_PARM_DESC(host_channels, "The number of host channel registers to use 1-16");
module_param_named(dev_endpoints, dwc_otg_module_params.dev_endpoints, int, 0444);
MODULE_PARM_DESC(dev_endpoints, "The number of endpoints in addition to EP0 available for device mode 1-15");
module_param_named(phy_type, dwc_otg_module_params.phy_type, int, 0444);
MODULE_PARM_DESC(phy_type, "0=Reserved 1=UTMI+ 2=ULPI");
module_param_named(phy_utmi_width, dwc_otg_module_params.phy_utmi_width, int, 0444);
MODULE_PARM_DESC(phy_utmi_width, "Specifies the UTMI+ Data Width 8 or 16 bits");
module_param_named(phy_ulpi_ddr, dwc_otg_module_params.phy_ulpi_ddr, int, 0444);
MODULE_PARM_DESC(phy_ulpi_ddr, "ULPI at double or single data rate 0=Single 1=Double");
module_param_named(phy_ulpi_ext_vbus, dwc_otg_module_params.phy_ulpi_ext_vbus, int, 0444);
MODULE_PARM_DESC(phy_ulpi_ext_vbus, "ULPI PHY using internal or external vbus 0=Internal");
module_param_named(i2c_enable, dwc_otg_module_params.i2c_enable, int, 0444);
MODULE_PARM_DESC(i2c_enable, "FS PHY Interface");
module_param_named(ulpi_fs_ls, dwc_otg_module_params.ulpi_fs_ls, int, 0444);
MODULE_PARM_DESC(ulpi_fs_ls, "ULPI PHY FS/LS mode only");
module_param_named(ts_dline, dwc_otg_module_params.ts_dline, int, 0444);
MODULE_PARM_DESC(ts_dline, "Term select Dline pulsing for all PHYs");
module_param_named(debug, g_dbg_lvl, int, 0444);
MODULE_PARM_DESC(debug, "");
#endif // CONFIG_IFX

/** @page "Module Parameters"
 *
 * The following parameters may be specified when starting the module.
 * These parameters define how the DWC_otg controller should be
 * configured.  Parameter values are passed to the CIL initialization
 * function dwc_otg_cil_init
 *
 * Example: <code>modprobe dwc_otg speed=1 otg_cap=1</code>
 *

 <table>
 <tr><td>Parameter Name</td><td>Meaning</td></tr>

 <tr>
 <td>otg_cap</td>
 <td>Specifies the OTG capabilities. The driver will automatically detect the
 value for this parameter if none is specified.
 - 0: HNP and SRP capable (default, if available)
 - 1: SRP Only capable
 - 2: No HNP/SRP capable
 </td></tr>

 <tr>
 <td>dma_enable</td>
 <td>Specifies whether to use slave or DMA mode for accessing the data FIFOs.
 The driver will automatically detect the value for this parameter if none is
 specified.
 - 0: Slave
 - 1: DMA (default, if available)
 </td></tr>

 <tr>
 <td>dma_burst_size</td>
 <td>The DMA Burst size (applicable only for External DMA Mode).
 - Values: 1, 4, 8 16, 32, 64, 128, 256 (default 32)
 </td></tr>

 <tr>
 <td>speed</td>
 <td>Specifies the maximum speed of operation in host and device mode. The
 actual speed depends on the speed of the attached device and the value of
 phy_type.
 - 0: High Speed (default)
 - 1: Full Speed
 </td></tr>

 <tr>
 <td>host_support_fs_ls_low_power</td>
 <td>Specifies whether low power mode is supported when attached to a Full
 Speed or Low Speed device in host mode.
 - 0: Don't support low power mode (default)
 - 1: Support low power mode
 </td></tr>

 <tr>
 <td>host_ls_low_power_phy_clk</td>
 <td>Specifies the PHY clock rate in low power mode when connected to a Low
 Speed device in host mode. This parameter is applicable only if
 HOST_SUPPORT_FS_LS_LOW_POWER is enabled.
 - 0: 48 MHz (default)
 - 1: 6 MHz
 </td></tr>

 <tr>
 <td>enable_dynamic_fifo</td>
 <td> Specifies whether FIFOs may be resized by the driver software.
 - 0: Use cC FIFO size parameters
 - 1: Allow dynamic FIFO sizing (default)
 </td></tr>

 <tr>
 <td>data_fifo_size</td>
 <td>Total number of 4-byte words in the data FIFO memory. This memory
 includes the Rx FIFO, non-periodic Tx FIFO, and periodic Tx FIFOs.
 - Values: 32 to 32768 (default 8192)

 Note: The total FIFO memory depth in the FPGA configuration is 8192.
 </td></tr>

 <tr>
 <td>dev_rx_fifo_size</td>
 <td>Number of 4-byte words in the Rx FIFO in device mode when dynamic
 FIFO sizing is enabled.
 - Values: 16 to 32768 (default 1064)
 </td></tr>

 <tr>
 <td>dev_nperio_tx_fifo_size</td>
 <td>Number of 4-byte words in the non-periodic Tx FIFO in device mode when
 dynamic FIFO sizing is enabled.
 - Values: 16 to 32768 (default 1024)
 </td></tr>

 <tr>
 <td>dev_perio_tx_fifo_size_n (n = 1 to 15)</td>
 <td>Number of 4-byte words in each of the periodic Tx FIFOs in device mode
 when dynamic FIFO sizing is enabled.
 - Values: 4 to 768 (default 256)
 </td></tr>

 <tr>
 <td>host_rx_fifo_size</td>
 <td>Number of 4-byte words in the Rx FIFO in host mode when dynamic FIFO
 sizing is enabled.
 - Values: 16 to 32768 (default 1024)
 </td></tr>

 <tr>
 <td>host_nperio_tx_fifo_size</td>
 <td>Number of 4-byte words in the non-periodic Tx FIFO in host mode when
 dynamic FIFO sizing is enabled in the core.
 - Values: 16 to 32768 (default 1024)
 </td></tr>

 <tr>
 <td>host_perio_tx_fifo_size</td>
 <td>Number of 4-byte words in the host periodic Tx FIFO when dynamic FIFO
 sizing is enabled.
 - Values: 16 to 32768 (default 1024)
 </td></tr>

 <tr>
 <td>max_transfer_size</td>
 <td>The maximum transfer size supported in bytes.
 - Values: 2047 to 65,535 (default 65,535)
 </td></tr>

 <tr>
 <td>max_packet_count</td>
 <td>The maximum number of packets in a transfer.
 - Values: 15 to 511 (default 511)
 </td></tr>

 <tr>
 <td>host_channels</td>
 <td>The number of host channel registers to use.
 - Values: 1 to 16 (default 12)

 Note: The FPGA configuration supports a maximum of 12 host channels.
 </td></tr>

 <tr>
 <td>dev_endpoints</td>
 <td>The number of endpoints in addition to EP0 available for device mode
 operations.
 - Values: 1 to 15 (default 6 IN and OUT)

 Note: The FPGA configuration supports a maximum of 6 IN and OUT endpoints in
 addition to EP0.
 </td></tr>

 <tr>
 <td>phy_type</td>
 <td>Specifies the type of PHY interface to use. By default, the driver will
 automatically detect the phy_type.
 - 0: Full Speed
 - 1: UTMI+ (default, if available)
 - 2: ULPI
 </td></tr>

 <tr>
 <td>phy_utmi_width</td>
 <td>Specifies the UTMI+ Data Width. This parameter is applicable for a
 phy_type of UTMI+. Also, this parameter is applicable only if the
 OTG_HSPHY_WIDTH cC parameter was set to "8 and 16 bits", meaning that the
 core has been configured to work at either data path width.
 - Values: 8 or 16 bits (default 16)
 </td></tr>

 <tr>
 <td>phy_ulpi_ddr</td>
 <td>Specifies whether the ULPI operates at double or single data rate. This
 parameter is only applicable if phy_type is ULPI.
 - 0: single data rate ULPI interface with 8 bit wide data bus (default)
 - 1: double data rate ULPI interface with 4 bit wide data bus
 </td></tr>

 <tr>
 <td>i2c_enable</td>
 <td>Specifies whether to use the I2C interface for full speed PHY. This
 parameter is only applicable if PHY_TYPE is FS.
 - 0: Disabled (default)
 - 1: Enabled
 </td></tr>

*/











extern void dwc_otg_enable_common_interrupts(dwc_otg_core_if_t *_core_if);
void init_devspd(dwc_otg_core_if_t *_core_if);


void dwc_otg_driver_reload(){
	int retval = 0;
	dwc_otg_device_t *otg_dev = lm_get_drvdata(_lmdev);
	int32_t	snpsid;

HDEBUG_S;

	if(otg_dev->pcd->ep0state == EP0_DISCONNECT)
		return;
	if(otg_dev->pcd->ep0state == EP0_STALL)
		return;

printk(KERN_INFO "RELOAD 01-01\n");
	DWC_DEBUGPL(DBG_ANY, "%s(%p)\n", __func__, otg_dev);
	if (otg_dev == NULL) {
		/* Memory allocation for the dwc_otg_device failed. */
HDEBUG_E(1);
		return;
	}





printk(KERN_INFO "RELOAD 01-02\n");

	//////////////////////////////////
	//   Unloading
	//////////////////////////////////
HDEBUG_C(1);
printk(KERN_INFO "RELOAD 01-01\n");
	if (otg_dev->pcd != NULL) {
printk(KERN_INFO "RELOAD 01-02\n");
//		if (otg_dev->pcd->driver)
//?			usb_gadget_unregister_driver( otg_dev->pcd->driver);

		if (otg_dev->core_if->dma_enable) {
printk(KERN_INFO "RELOAD 01-03\n");
			dma_free_coherent (NULL,
			                   sizeof (*otg_dev->pcd->setup_pkt) * 5,
			                   otg_dev->pcd->setup_pkt,
			                   otg_dev->pcd->setup_pkt_dma_handle);
printk(KERN_INFO "RELOAD 01-04\n");
			dma_free_coherent (NULL,
			                   sizeof (uint16_t),
			                   otg_dev->pcd->status_buf,
			                   otg_dev->pcd->status_buf_dma_handle);
printk(KERN_INFO "RELOAD 01-05\n");
		}
		else {
printk(KERN_INFO "RELOAD 01-06\n");
			kfree (otg_dev->pcd->setup_pkt);
printk(KERN_INFO "RELOAD 01-07\n");
			kfree (otg_dev->pcd->status_buf);
printk(KERN_INFO "RELOAD 01-08\n");
		}
	}


	//////////////////////////////////
	//   Loading
	//////////////////////////////////
//	otg_dev->core_if = dwc_otg_cil_init( otg_dev->base, &dwc_otg_module_params);
	{
		otg_dev->core_if->dev_if->speed = 0; 					// unknown
		otg_dev->core_if->srp_success = 0;
		otg_dev->core_if->srp_timer_started = 0;
//?		otg_dev->core_if->phy_init_done=0;
//?		otg_dev->core_if->dma_enable=0;
		otg_dev->core_if->queuing_high_bandwidth=0;
		otg_dev->core_if->op_state=0;
		otg_dev->core_if->restart_hcd_on_session_req=0;
printk(KERN_INFO "RELOAD 02-01\n");
	}
printk(KERN_INFO "RELOAD 03-01\n");
	dwc_otg_disable_global_interrupts( otg_dev->core_if );
printk(KERN_INFO "RELOAD 04-01\n");
//	dwc_otg_core_init( otg_dev->core_if );
	{
		gahbcfg_data_t ahbcfg = {.d32 = 0};
		gusbcfg_data_t usbcfg = {.d32 = 0};
		gi2cctl_data_t i2cctl = {.d32 = 0};
		int i = 0;
printk(KERN_INFO "RELOAD 05-01\n");
		dwc_otg_core_reset( otg_dev->core_if );
printk(KERN_INFO "RELOAD 06-01\n");
		// Now do PHY ?
#if 1
//PHY
		/* This programming sequence needs to happen in FS mode before any other
		 * programming occurs */
		if ((otg_dev->core_if->core_params->speed == DWC_SPEED_PARAM_FULL) &&
		    (otg_dev->core_if->core_params->phy_type == DWC_PHY_TYPE_PARAM_FS)) {
			/* If FS mode with FS PHY */

			/* core_init() is now called on every switch so only call the
			 * following for the first time through. */
			if (!otg_dev->core_if->phy_init_done) {
				otg_dev->core_if->phy_init_done = 1;
				DWC_DEBUGPL(DBG_CIL, "FS_PHY detected\n");
				usbcfg.d32 = dwc_read_reg32(&otg_dev->core_if->core_global_regs->gusbcfg);
				usbcfg.b.physel = 1;
				dwc_write_reg32 (&otg_dev->core_if->core_global_regs->gusbcfg, usbcfg.d32);

				/* Reset after a PHY select */
				dwc_otg_core_reset( otg_dev->core_if );
			}

			/* Program DCFG.DevSpd or HCFG.FSLSPclkSel to 48Mhz in FS.  Also
			 * do this on HNP Dev/Host mode switches (done in dev_init and
			 * host_init). */
			init_devspd(otg_dev->core_if);

			if (otg_dev->core_if->core_params->i2c_enable) {
				DWC_DEBUGPL(DBG_CIL, "FS_PHY Enabling I2c\n");
				/* Program GUSBCFG.OtgUtmifsSel to I2C */
				usbcfg.d32 = dwc_read_reg32(&otg_dev->core_if->core_global_regs->gusbcfg);
				usbcfg.b.otgutmifssel = 1;
				dwc_write_reg32 (&otg_dev->core_if->core_global_regs->gusbcfg, usbcfg.d32);

				/* Program GI2CCTL.I2CEn */
				i2cctl.d32 = dwc_read_reg32(&otg_dev->core_if->core_global_regs->gi2cctl);
				i2cctl.b.i2cdevaddr = 1;
				i2cctl.b.i2cen = 0;
				dwc_write_reg32 (&otg_dev->core_if->core_global_regs->gi2cctl, i2cctl.d32);
				i2cctl.b.i2cen = 1;
				dwc_write_reg32 (&otg_dev->core_if->core_global_regs->gi2cctl, i2cctl.d32);
			}
		} /* endif speed == DWC_SPEED_PARAM_FULL */
		else {
			/* High speed PHY. */
			if (!otg_dev->core_if->phy_init_done) {
				otg_dev->core_if->phy_init_done = 1;
				/* HS PHY parameters.  These parameters are preserved
				 * during soft reset so only program the first time.  Do
				 * a soft reset immediately after setting phyif.  */
				usbcfg.b.ulpi_utmi_sel = otg_dev->core_if->core_params->phy_type;
	#ifdef USE_ORIG
				if (usbcfg.b.ulpi_utmi_sel == 1) { // winder
	#else
				if (usbcfg.b.ulpi_utmi_sel == 2) { // winder
	#endif
					/* ULPI interface */
					usbcfg.b.phyif = 0;
					usbcfg.b.ddrsel = otg_dev->core_if->core_params->phy_ulpi_ddr;
				} else {
					/* UTMI+ interface */
					if (otg_dev->core_if->core_params->phy_utmi_width == 16) {
						usbcfg.b.phyif = 1;
					} else {
						usbcfg.b.phyif = 0;
					}
				}
				dwc_write_reg32( &otg_dev->core_if->core_global_regs->gusbcfg, usbcfg.d32);

				/* Reset after setting the PHY parameters */
				dwc_otg_core_reset( otg_dev->core_if );
			}
		}

		if ((otg_dev->core_if->hwcfg2.b.hs_phy_type == 2) &&
		    (otg_dev->core_if->hwcfg2.b.fs_phy_type == 1) &&
		    (otg_dev->core_if->core_params->ulpi_fs_ls)){
			DWC_DEBUGPL(DBG_CIL, "Setting ULPI FSLS\n");
			usbcfg.d32 = dwc_read_reg32(&otg_dev->core_if->core_global_regs->gusbcfg);
			usbcfg.b.ulpi_fsls = 1;
			usbcfg.b.ulpi_clk_sus_m = 1;
			dwc_write_reg32(&otg_dev->core_if->core_global_regs->gusbcfg, usbcfg.d32);
		}
		else {
			usbcfg.d32 = dwc_read_reg32(&otg_dev->core_if->core_global_regs->gusbcfg);
			usbcfg.b.ulpi_fsls = 0;
			usbcfg.b.ulpi_clk_sus_m = 0;
			dwc_write_reg32(&otg_dev->core_if->core_global_regs->gusbcfg, usbcfg.d32);
		}
#endif

		/* Enable common interrupts */
printk(KERN_INFO "RELOAD 06-01\n");
		dwc_otg_enable_common_interrupts( otg_dev->core_if );
printk(KERN_INFO "RELOAD 07-01\n");
		otg_dev->core_if->op_state = B_PERIPHERAL;
printk(KERN_INFO "RELOAD 08-01\n");
//		dwc_otg_core_dev_init( otg_dev->core_if );
		{
			dwc_otg_core_global_regs_t *global_regs = otg_dev->core_if->core_global_regs;
			dwc_otg_dev_if_t *dev_if = otg_dev->core_if->dev_if;
			dwc_otg_core_params_t *params = otg_dev->core_if->core_params;

			grstctl_t resetctl = { .d32=0 };
			dcfg_data_t dcfg = {.d32 = 0};
			int i;
			uint32_t rx_fifo_size;
			fifosize_data_t nptxfifosize;
#ifdef USE_PERIODIC_EP
			fifosize_data_t ptxfifosize;
#endif

printk(KERN_INFO "RELOAD 09-01\n");
			/* Restart the Phy Clock */
			dwc_write_reg32(otg_dev->core_if->pcgcctl, 0);
printk(KERN_INFO "RELOAD 09-02\n");
			init_devspd(otg_dev->core_if);
printk(KERN_INFO "RELOAD 09-03\n");
			/* Device configuration register */
			dcfg.d32 = dwc_read_reg32( &dev_if->dev_global_regs->dcfg);
			dcfg.b.perfrint = DWC_DCFG_FRAME_INTERVAL_80;
			dwc_write_reg32( &dev_if->dev_global_regs->dcfg, dcfg.d32 );
printk(KERN_INFO "RELOAD 09-04\n");
			/* Configure data FIFO sizes */
			if ( otg_dev->core_if->hwcfg2.b.dynamic_fifo && params->enable_dynamic_fifo ) {
printk(KERN_INFO "RELOAD 09-05\n");
				DWC_DEBUGPL(DBG_CIL, "Total FIFO Size=%d\n", otg_dev->core_if->total_fifo_size);
				DWC_DEBUGPL(DBG_CIL, "Rx FIFO Size=%d\n", params->dev_rx_fifo_size);
				DWC_DEBUGPL(DBG_CIL, "NP Tx FIFO Size=%d\n", params->dev_nperio_tx_fifo_size);
				/* Rx FIFO */
				DWC_DEBUGPL(DBG_CIL, "initial grxfsiz=%08x\n", dwc_read_reg32(&global_regs->grxfsiz));
				rx_fifo_size = params->dev_rx_fifo_size;
				dwc_write_reg32( &global_regs->grxfsiz, rx_fifo_size );
				DWC_DEBUGPL(DBG_CIL, "new grxfsiz=%08x\n", dwc_read_reg32(&global_regs->grxfsiz));

				/* Non-periodic Tx FIFO */
				DWC_DEBUGPL(DBG_CIL, "initial gnptxfsiz=%08x\n", dwc_read_reg32(&global_regs->gnptxfsiz));
				nptxfifosize.b.depth  = params->dev_nperio_tx_fifo_size;
				nptxfifosize.b.startaddr = params->dev_rx_fifo_size;
				dwc_write_reg32( &global_regs->gnptxfsiz, nptxfifosize.d32 );
				DWC_DEBUGPL(DBG_CIL, "new gnptxfsiz=%08x\n", dwc_read_reg32(&global_regs->gnptxfsiz));
#ifdef USE_PERIODIC_EP
				/**@todo NGS: Fix Periodic FIFO Sizing! */
				/*
				 * Periodic Tx FIFOs These FIFOs are numbered from 1 to 15.
				 * Indexes of the FIFO size module parameters in the
				 * dev_perio_tx_fifo_size array and the FIFO size registers in
				 * the dptxfsiz array run from 0 to 14.
				 */
				/** @todo Finish debug of this */
				ptxfifosize.b.startaddr = nptxfifosize.b.startaddr + nptxfifosize.b.depth;
				for (i=0; i < dev_if->num_perio_eps; i++) {
					ptxfifosize.b.depth = params->dev_perio_tx_fifo_size[i];
					DWC_DEBUGPL(DBG_CIL, "initial dptxfsiz[%d]=%08x\n", i,
					     dwc_read_reg32(&global_regs->dptxfsiz[i]));
					dwc_write_reg32( &global_regs->dptxfsiz[i], ptxfifosize.d32 );
					DWC_DEBUGPL(DBG_CIL, "new dptxfsiz[%d]=%08x\n", i,
					     dwc_read_reg32(&global_regs->dptxfsiz[i]));
					ptxfifosize.b.startaddr += ptxfifosize.b.depth;
				}
#endif
			}
			/* Flush the FIFOs */
printk(KERN_INFO "RELOAD 09-06\n");
			dwc_otg_flush_tx_fifo(otg_dev->core_if, 0x10); /* all Tx FIFOs */
			dwc_otg_flush_tx_fifo(otg_dev->core_if, 0x0); /* all Tx FIFOs */
printk(KERN_INFO "RELOAD 09-07\n");
			dwc_otg_flush_rx_fifo(otg_dev->core_if);
printk(KERN_INFO "RELOAD 09-08\n");

			/* Flush the Learning Queue. */
			resetctl.b.intknqflsh = 1;
			dwc_write_reg32( &global_regs->grstctl, resetctl.d32);
printk(KERN_INFO "RELOAD 09-09\n");

			/* Clear all pending Device Interrupts */
			dwc_write_reg32( &dev_if->dev_global_regs->diepmsk, 0 );
printk(KERN_INFO "RELOAD 09-10\n");
			dwc_write_reg32( &dev_if->dev_global_regs->doepmsk, 0 );
printk(KERN_INFO "RELOAD 09-11\n");
			dwc_write_reg32( &dev_if->dev_global_regs->daint, 0xFFFFFFFF );
printk(KERN_INFO "RELOAD 09-12\n");
			dwc_write_reg32( &dev_if->dev_global_regs->daintmsk, 0 );
printk(KERN_INFO "RELOAD 09-13\n");

			for (i=0; i < dev_if->num_eps; i++) {
				depctl_data_t depctl;
				depctl.d32 = dwc_read_reg32(&dev_if->in_ep_regs[i]->diepctl);
				if (depctl.b.epena) {
					depctl.d32 = 0;
					depctl.b.epdis = 1;
					depctl.b.snak = 1;
				} else {
					depctl.d32 = 0;
				}
				dwc_write_reg32( &dev_if->in_ep_regs[i]->diepctl, depctl.d32);
printk(KERN_INFO "RELOAD 09-14\n");

				depctl.d32 = dwc_read_reg32(&dev_if->out_ep_regs[i]->doepctl);
				if (depctl.b.epena) {
					depctl.d32 = 0;
					depctl.b.epdis = 1;
					depctl.b.snak = 1;
				} else {
					depctl.d32 = 0;
				}
				dwc_write_reg32( &dev_if->out_ep_regs[i]->doepctl, depctl.d32);
				dwc_write_reg32( &dev_if->in_ep_regs[i]->dieptsiz, 0);
				dwc_write_reg32( &dev_if->out_ep_regs[i]->doeptsiz, 0);
				dwc_write_reg32( &dev_if->in_ep_regs[i]->diepdma, 0);
				dwc_write_reg32( &dev_if->out_ep_regs[i]->doepdma, 0);
				dwc_write_reg32( &dev_if->in_ep_regs[i]->diepint, 0xFF);
				dwc_write_reg32( &dev_if->out_ep_regs[i]->doepint, 0xFF);
printk(KERN_INFO "RELOAD 09-15\n");
			}
			dwc_otg_enable_device_interrupts( otg_dev->core_if );
printk(KERN_INFO "RELOAD 09-16\n");
		}
printk(KERN_INFO "RELOAD 09-17\n");
	}
//	retval = dwc_otg_pcd_init(otg_dev);
	{
		static char pcd_name[] = "dwc_otg_pcd";
		dwc_otg_pcd_t *pcd=otg_dev->pcd;
		int retval = 0;

		pcd->ep0_pending=0;
		pcd->request_config=0;
		pcd->remote_wakeup_enable=0;
		pcd->b_hnp_enable=0;
		pcd->a_hnp_support=0;
		pcd->a_alt_hnp_support=0;
		pcd->request_pending=0;
		//pcd->srp_timer=0;
		pcd->test_mode=0;

printk(KERN_INFO "RELOAD 10-01\n");
//?		pcd->driver = 0;
//		dwc_otg_pcd_reinit( pcd );
		{
			static const char * names[] = {
			    "ep0", "ep1", "ep2", "ep3",
			    "ep4", "ep5", "ep6", "ep7",
			    "ep8", "ep9", "ep10", "ep11",
			    "ep12", "ep13", "ep14", "ep15",
			};
			uint32_t i;
printk(KERN_INFO "RELOAD 10-02\n");
			INIT_LIST_HEAD (&pcd->gadget.ep_list);
			pcd->gadget.ep0 = &pcd->ep [0].ep;
			pcd->gadget.speed = USB_SPEED_UNKNOWN;
printk(KERN_INFO "RELOAD 10-03\n");
			INIT_LIST_HEAD (&pcd->gadget.ep0->ep_list);
printk(KERN_INFO "RELOAD 10-04\n");
			for (i = 0; i < pcd->num_eps; i++) {
				dwc_otg_pcd_ep_t *ep = &pcd->ep[i];
				/* Init EP structure */
				ep->desc = 0;
				ep->pcd = pcd;
				ep->stopped = 1;

				/* Init DWC ep structure */
				ep->dwc_ep.num = i;
				ep->dwc_ep.active = 0;
				ep->dwc_ep.tx_fifo_num = 0;

				/* Control until ep is actvated */
				ep->dwc_ep.type = DWC_OTG_EP_TYPE_CONTROL;
				ep->dwc_ep.maxpacket = MAX_PACKET_SIZE;
				ep->dwc_ep.dma_addr = 0;
				ep->dwc_ep.start_xfer_buff = 0;
				ep->dwc_ep.xfer_buff = 0;
				ep->dwc_ep.xfer_len = 0;
				ep->dwc_ep.xfer_count = 0;
				ep->dwc_ep.sent_zlp = 0;
				ep->dwc_ep.total_len = 0;
				ep->queue_sof = 0;

				/* Init the usb_ep structure. */
				/**
				 * @todo NGS: Add direction to EP, based on contents
				 * of HWCFG1.  Need a copy of HWCFG1 in pcd structure?
				 * sprintf( ";r
				 */
				ep->ep.name = names[i];
				//ep->ep.ops = &dwc_otg_pcd_ep_ops;
				/**
				 * @todo NGS: What should the max packet size be set to
				 * here?  Before EP type is set?
				 */
				ep->ep.maxpacket = MAX_PACKET_SIZE;
printk(KERN_INFO "RELOAD 10-05\n");
				list_add_tail (&ep->ep.ep_list, &pcd->gadget.ep_list);
				INIT_LIST_HEAD (&ep->queue);
			}
			/* remove ep0 from the list.  There is a ep0 pointer.*/
printk(KERN_INFO "RELOAD 10-06\n");
			list_del_init (&pcd->ep[0].ep.ep_list);
printk(KERN_INFO "RELOAD 10-07\n");

			pcd->ep0state = EP0_DISCONNECT;
			pcd->ep[0].ep.maxpacket = MAX_EP0_SIZE;
			pcd->ep[0].dwc_ep.maxpacket = MAX_EP0_SIZE;
			pcd->ep[0].dwc_ep.type = DWC_OTG_EP_TYPE_CONTROL;
		}
		/*
		 * Initialize the DMA buffer for SETUP packets
		 */
		if (otg_dev->core_if->dma_enable) {
printk(KERN_INFO "RELOAD 11-01\n");
			pcd->setup_pkt = dma_alloc_coherent (NULL, sizeof (*pcd->setup_pkt) * 5, &pcd->setup_pkt_dma_handle, 0);
printk(KERN_INFO "RELOAD 11-02\n");
			pcd->status_buf = dma_alloc_coherent (NULL, sizeof (uint16_t), &pcd->status_buf_dma_handle, 0);
printk(KERN_INFO "RELOAD 11-03\n");
		}
		else {
printk(KERN_INFO "RELOAD 11-04\n");
			pcd->setup_pkt = kmalloc (sizeof (*pcd->setup_pkt) * 5, GFP_KERNEL);
printk(KERN_INFO "RELOAD 11-05\n");
			pcd->status_buf = kmalloc (sizeof (uint16_t), GFP_KERNEL);
printk(KERN_INFO "RELOAD 11-06\n");
		}
	}
	dwc_otg_enable_global_interrupts( otg_dev->core_if );

}
